/* start of file Eigen.cpp ------------------------------------------------ */

/* <summary>
	SUBROUTINE Eigen - Computes Eigenvalues and Eigenvectors of
	an n by n complex matrix a
   </summary>
*/

#include "stdafx.h"
#include "Eigen.h"

using namespace std;

void DelC(complex<float> x[][maxA], int nr, int *nc, int c);
void DelR(complex<float> x[], int *nr, int r);
void CharPoly(complex<float> a[][maxA], int n, complex<float> p[]);
void PolRoots(complex<float> p[], int n, bool polish, complex<float> r[]);
void CSVD(complex<float> a[][maxA], int m, int n, int p, int nu, int nv,
			float s[], complex<float> u[][maxA], complex<float> v[][maxA]);

void Eigen(complex<float> a[][maxA], int n, complex<float> va[], complex<float> ve[][maxA], bool herm, bool *error)
/* va = Eigenvalues and ve = Eigenvectors
   of an n by n matrix a, Hermitian if herm == true, n > 1
*/
{
	*error = false;
	complex<float> cZero = complex<float>(0.0F, 0.0F);
	complex<float> cOne = complex<float>(1.0F, 0.0F);
	complex<float> p[maxA + 1];
	CharPoly(a, n, p);				/* p = characteristic polynomial of square matrix a */
	PolRoots(p, n+1, true, va);		/* va = All n of the roots of polynomial p */
	int i, j, k, ri, ci, f = 0;
	if (herm)
		for (i = 0; i < n; i++)
			va[i] = complex<float>(real(va[i]), 0.0F);
	int vaRs = n;
	int veCs = n;
	for (i = n - 2; i >= 0; i--)	/* delete roots that are duplicates */
		if (va[i] == va[i + 1])
		{
			DelR(va, &vaRs, i);
		}
	complex<float> aa[maxA][maxA];
	complex<float> u[maxA][maxA];
	float w[maxA];
	complex<float> v[maxA][maxA];
	complex<float> di;
	for (i = 0; i < n; i++)
	{
		for (j = 0; j < n; j++)
			for (k = 0; k < n; k++)			/* u = a */
				aa[j][k] = a[j][k];
		for (j = 0; j < n; j++)				/* subtract i-th root from all diagonal elements of u */
			aa[j][j] -= va[i];
		int p = 0;
		CSVD(aa, n, n, p, n, n, w, u, v);	/* Decompose u to (u, w, v) */
		int vCs = n;
		for (j = n - 1; j > 0; j--)
		{
			if (w[0] != 0.0F && w[j] / w[0] <= 5.0E-7F)	/* how small is too small */
				w[j] = 0;
		}
		for (j = n - 1; j >= 0; j--)		/* Delete columns of v for w != 0 */
		{
			if (w[j] != 0.0F)				/* w is real */
				DelC(v, n, &vCs, j);
		}
		for (ci = 0; ci < vCs; ci++)		/* concat columns of v to ve */
		{
			if (f < n)	/* for safety */
			{
				for (ri = 0; ri < n; ri++)
				{
					ve[ri][f] = v[ri][ci];
				}
				f++;
			}
			else
				break;
		}
	}
	for (ci = n - 1; ci >= 0; ci--)			/* for all columns of ve */
	{
		di = cZero;
		for (ri = n - 1; ri >= 0; ri--)		/* find last non-zero element in this column */
		{
			if (ve[ri][ci] != cZero)
			{
				di = ve[ri][ci];
				break;
			}
		}
		if (di == cZero)								/* if column is all zeros */
			DelC(ve, n, &veCs, ci);					/* Delete columns of ve with all zeros */
		else
		{
			for (ri = 0; ri < n; ri++)				/* Make last non-zero element equal to one */
			{
				if (ve[ri][ci] == di)
					ve[ri][ci] = cOne;
				else
					ve[ri][ci] /= di;
			}
		}
	}
	for (ci = n - 1; ci > 0; ci--)
	{
		int d = 0;
		for (ri = 0; ri < n; ri++)
		{
			if (ve[ri][ci] == ve[ri][ci - 1])
				d++;
			else
				break;
		}
		if (d == n)
		{
			DelC(ve, n, &veCs, ci);
			*error = true;					/* Not all eigenvectors found */
		}
	}
} /* Eigen */

void DelC(complex<float> x[][maxA], int nr, int *nc, int c)
/* Delete column c from nr by nc matrix x (0 < c < nc - 1)
*/
{
	if (c < 0 || c >= *nc - 1 || *nc <= 1)
		return;
	for (int ri = 0; ri < nr; ri++)
		for (int ci = c + 1; ci < *nc; ci++)
			x[ri][ci - 1] = x[ri][ci];
	(*nc)--;
}

void DelR(complex<float> x[], int *nr, int r)
/* Delete row r from nr row vector x (0 < r < nr - 1)
*/
{
	if (r < 0 || r >= *nr - 1 || *nr <= 1)
		return;
	for (int ri = r + 1; ri < *nr; ri++)
		x[ri - 1] = x[ri];
	(*nr)--;
}

/* end of file Eigen.cpp -------------------------------------------- */
