/* start of file PolRoots.cpp --------------------------------------------- */

/* <summary>
	SUBROUTINE PolRoots - Computes the n roots on the complex polynomial
	p = p[0], p[1],..., p[n]
   </summary>
*/

#include "stdafx.h"
#include "PolRoots.h"

using namespace std;

void PolNorm(complex<float> a[], int *n);
void Laguer(complex<float> a[], int n, int *its, complex<float> *x);
void ClearSmall(complex<float> x);

void PolRoots(complex<float> a[], int n, bool polish, complex<float> roots[])
/* roots = The m = n-1 roots of complex polynomial a = a[0], a[1],..., a[m]
*/
/* Adapted from http://www.library.cornell.edu/nr/bookfpdf/f9-5.pdf
   The coefficient a[m] is the constant term, while a[0] is the
   coefficient of the highest power of x. This routine calls Laguer in
   succession for each root, performs the deflation, optionally
   polishes the roots, if polish = true, by the same Laguerre method
   -- if you are not going to polish in some other way -- and finally
   sorts the roots by their real parts.
*/
{
	int m = n - 1;			/* n = size of polynomial, m = degree of polynomial */
	int i, j, jj, its;
	complex<float> aa[maxN];
	complex<float> cZero = complex<float>(0.0F, 0.0F);
	for (i = 0; i < n; i++)	/* aa = a */
		aa[i] = a[i];
	int aan = n;
	PolNorm(aa, &aan);
	complex<float> ad[maxN];
	complex<float> x;
	complex<float> b;
	complex<float> c;
	for (i = 0; i < n; i++)	/* ad = aa */
		ad[i] = aa[i];
	int adn = n;
	for (j = 0; j < m; j++)		/* Loop over each root to be found. */
	{
		/* Start at zero to favor convergence to smallest remaining root */
		x = cZero;
		Laguer(ad, adn, &its, &x);	/* Find the root */
		ClearSmall(x);				/* clear small real or imaginary part */
		roots[j] = x;				/* save the root */
		b = ad[0];					/* Forward deflation. */
		for (jj = 1; jj <= m - j; jj++)
		{
			c = ad[jj];
			ad[jj - 1] = b;
			b = x * b + c;
		}
		adn--;	/* fix ad to have one less row */
	}
	if (polish)
	{
		/* Polish the roots using the undeflated coefficients */
		for (j = 0; j < m; j++)
			Laguer(a, n, &its, &roots[j]);
	}
	for (j = 1; j < m; j++)	/* Sort roots by straight insertion */
	{
		x = roots[j];
		for (i = j - 1; i >= 0; i--)
		{
			if (real(roots[i]) < real(x))	/* if (roots[i] <= x) */
				break;
			if ((real(roots[i]) == real(x)) && (imag(roots[i]) <= imag(x)))
				break;
			roots[i + 1] = roots[i];
		}
		roots[i + 1] = x;
	}
} /* PolRoots */

void PolNorm(complex<float> a[], int *n)
/* a = norm of polynomial a, leading coeff. not zero
*/
{
	complex<float> cZero = complex<float>(0.0F, 0.0F);
	if (a[0] != cZero)
		return;
	int nz = 0;
	int r;
	for (r = 0; r < *n - 1; r++)
		if (a[r] == cZero)
			nz++;
		else
			break;
	for (r = nz; r < *n; r++)
		a[r - nz] = a[r];
	*n -= nz;
} /* PolNorm */

void ClearSmall(complex<float> x)
{
	if (real(x) != 0.0F && imag(x) / real(x) <= 5.0E-7F)
	{
		x = complex<float>(real(x), 0.0F);
		return;
	}
	if (imag(x) != 0.0F && real(x) / imag(x) <= 5.0E-7F)
		x = complex<float>(0.0F, imag(x));
} /* ClearSmall */

/* end of file PolRoots.cpp ----------------------------------------------- */

